#pragma once

#include<vector>

using std::size_t;

size_t upper_power_of_two(size_t v)
{
    v--;
    v |= v >> 1;
    v |= v >> 2;
    v |= v >> 4;
    v |= v >> 8;
    v |= v >> 16;
    v++;
    return v;
}

enum { SMART_MAX = 0, NAIVE_MAX = 1 };

template<typename T, int options = SMART_MAX>
struct buffer {

  std::vector<T> data;
  size_t potsize;
  bool tree_is_dirty;
  size_t tree_dirty_start;
  size_t tree_dirty_last;

  buffer() : potsize(0), tree_is_dirty(false) {}

  const T& element(size_t i) const { return data[storage(i)]; }
  T& element(size_t i) { return data[storage(i)]; }

  T max(size_t first, size_t count) {
    T result = 0;
    if (options == NAIVE_MAX) {
      size_t start = storage(first);
      size_t last = storage(first + count - 1);
      for (size_t i = start; i <= last; i++)
        result = std::max(result, data[i]);
    } else {
      size_t start = storage(first);
      size_t last = storage(first + count - 1);
      update_tree_if_needed(start, last);
      while (true) {
        if (last <= start) {
          if (last < start)
            return result;
          result = std::max(result, data[start]);
          return result;
        }
        if (is_right(start)) {
          result = std::max(result, data[start]);
          start = right_neighbour(start);
        }
        if (!is_right(last)) {
          result = std::max(result, data[last]);
          last = left_neighbour(last);
        }
        start = parent(start);
        last = parent(last);
      }
    }
    return result;
  }

  void bufferData(const std::vector<T>& d) {
    bufferData(d.data(), d.size());
  }

  void bufferData(const T* ptr, size_t size) {
    potsize = upper_power_of_two(size);
    data.resize(2 * potsize);
    tree_is_dirty = false;
    bufferSubData(0, ptr, size);
  }

  void bufferSubData(size_t pos, const std::vector<T>& d) {
    bufferSubData(pos, d.data(), d.size());
  }

  void bufferSubData(size_t pos, const T* ptr, size_t size) {
    for (size_t i = 0; i < size; i++)
      element(pos + i) = ptr[i];

    invalidate(storage(pos), storage(pos + size - 1));
  }

private:

  size_t storage(size_t user_index) const {
    return potsize + user_index;
  }

  static size_t parent(size_t i) {
    return i >> 1;
  }

  static bool is_right(size_t i) {
    return i & 1;
  }

  static bool is_left(size_t i) {
    return !is_right(i);
  }

  static size_t left_child(size_t i) {
    return i << 1;
  }

  static size_t right_child(size_t i) {
    return sibling(left_child(i));
  }

  static size_t left_neighbour(size_t i) {
    return i - 1;
  }

  static size_t right_neighbour(size_t i) {
    return i + 1;
  }

  static size_t sibling(size_t i) {
    return i ^ 1;
  }

  void update_tree_if_needed(size_t start, size_t last) {
    if (tree_is_dirty &&
        start <= tree_dirty_last &&
        last >= tree_dirty_start)
    {
      update_tree(tree_dirty_start, tree_dirty_last);
    }
  }

  void invalidate(size_t start, size_t last) {
    if (tree_is_dirty) {
      tree_dirty_start = std::min(tree_dirty_start, start);
      tree_dirty_last = std::min(tree_dirty_last, last);
    } else {
      tree_is_dirty = true;
      tree_dirty_start = start;
      tree_dirty_last = last;
    }
  }

  void update_tree(size_t start, size_t last) {
    tree_is_dirty = false;
    if (options == NAIVE_MAX)
      return;
    if (!potsize)
      return;

    while (true) {
      if (start == last) {
        size_t startparent = parent(start);
        while (startparent) {
          data[startparent] = std::max(data[start], data[sibling(start)]);
          start = startparent;
          startparent = parent(start);
        }
        return;
      }
      start = parent(start);
      last = parent(last);
      size_t child = left_child(start);
      for (size_t parent = start; parent <= last; parent = right_neighbour(parent))
      {
        T a = data[child];
        child = right_neighbour(child);
        T b = data[child];
        child = right_neighbour(child);
        data[parent] = std::max(a, b);
      }
    }
  }

};
